/*
 * Copyright (C) 2020  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef LOG_ENTRY_HH_C42FFA9E17BE44D48FAB7F5FB13A9E02
#define LOG_ENTRY_HH_C42FFA9E17BE44D48FAB7F5FB13A9E02

#include "libpg/pg_rw_transaction.hh"
#include "libpg/util/strong_type.hh"

#include <boost/asio/ip/address.hpp>
#include <boost/date_time/posix_time/ptime.hpp>

#include <string>
#include <vector>

namespace Fred {
namespace Logger {

namespace LogEntry {

using TimeBegin = LibPg::Util::StrongType<boost::posix_time::ptime, struct TimeBeginTag_>;
using TimeEnd = LibPg::Util::StrongType<boost::posix_time::ptime, struct TimeEndTag_>;
using SourceIp = LibPg::Util::StrongType<boost::asio::ip::address, struct SourceIpTag_>;
using ResultCode = LibPg::Util::StrongType<unsigned long long, struct ResultCodeTag_>;
using ResultName = LibPg::Util::StrongType<std::string, struct ResultNameTag_>;

using SessionId = LibPg::Util::StrongType<unsigned long long, struct SessionIdTag_>;
using SessionLoginDate = LibPg::Util::StrongType<boost::posix_time::ptime, struct SessionLoginDateTag_>;
using SessionLogoutDate = LibPg::Util::StrongType<boost::posix_time::ptime, struct SessionLogoutDateTag_>;
using UserName = LibPg::Util::StrongType<std::string, struct UserNameTag_>;
using UserId = LibPg::Util::StrongType<unsigned long long, struct UserIdTag_>;

using ServiceId = LibPg::Util::StrongType<int, struct ServiceIdTag_>;
using ServiceName = LibPg::Util::StrongType<std::string, struct ServiceNameTag_>;
using IsMonitoring = LibPg::Util::StrongType<bool, struct IsMonitoringTag_>;
using LogEntryId = LibPg::Util::StrongType<unsigned long long, struct LogEntryIdTag_>;
using Content = LibPg::Util::StrongType<std::string, struct ContentTag_>;
using RawRequest = LibPg::Util::StrongType<std::string, struct RawRequestTag_>;
using RawResponse = LibPg::Util::StrongType<std::string, struct RawResponseTag_>;
using IsResponse = LibPg::Util::StrongType<bool, struct IsResponseTag_>;

using PropertyId = LibPg::Util::StrongType<unsigned long long, struct PropertyIdTag_>;
using PropertyNameId = LibPg::Util::StrongType<int, struct PropertyNameIdTag_>;
using PropertyType = LibPg::Util::StrongType<std::string, struct PropertyTypeTag_>;
using PropertyValue = LibPg::Util::StrongType<std::string, struct PropertyValueTag_>;
using IsOutput = LibPg::Util::StrongType<bool, struct IsOutputTag_>;

using LogEntryObjectTypeName = LibPg::Util::StrongType<std::string, struct LogEntryObjectTypeNameTag_>;
using LogEntryObjectTypeId = LibPg::Util::StrongType<unsigned int, struct LogEntryObjectTypeIdTag_>;
using ObjectId = LibPg::Util::StrongType<int, struct ObjectIdTag_>;
using ObjectIdent = LibPg::Util::StrongType<std::string, struct ObjectIdentTag_>;

using LogEntryTypeName = LibPg::Util::StrongType<std::string, struct LogEntryTypeNameTag_>;
using LogEntryTypeId = LibPg::Util::StrongType<unsigned int, struct LogEntryTypeIdTag_>;

using CorbaImplCompatibility = LibPg::Util::StrongType<bool, struct CorbaImplCompatibilityTag_>;

bool is_valid(const ServiceName& _service_name);

void insert_log_entry_data(
        const LibPg::PgRwTransaction& _rw_transaction,
        const TimeBegin& _time_begin,
        const ServiceId& _service_id,
        const IsMonitoring& _is_monitoring,
        const LogEntryId& _log_entry_id,
        const Content& _log_entry_content,
        const IsResponse& _is_response);

PropertyId insert_log_entry_property_as_parent(
        const LibPg::PgRwTransaction& _rw_transaction,
        const TimeBegin& _time_begin,
        const ServiceId& _service_id,
        const IsMonitoring& _is_monitoring,
        const LogEntryId& _log_entry_id,
        const PropertyNameId& _property_name_id,
        const PropertyValue& _property_value,
        const IsOutput& _is_output);

void insert_log_entry_property_as_child(
        const LibPg::PgRwTransaction& _rw_transaction,
        const TimeBegin& _time_begin,
        const ServiceId& _service_id,
        const IsMonitoring& _is_monitoring,
        const LogEntryId& _log_entry_id,
        const PropertyNameId& _property_name_id,
        const PropertyValue& _property_value,
        const IsOutput& _is_output,
        const PropertyId& _parent_id);

void insert_object_reference(
        const LibPg::PgRwTransaction& _rw_transaction,
        const TimeBegin& _time_begin,
        const ServiceId& _service_id,
        const IsMonitoring& _is_monitoring,
        const LogEntryId& _log_entry_id,
        const LogEntryObjectTypeId& _log_entry_object_type_id,
        const ObjectIdent& _object_id);

std::tuple<ObjectId, boost::optional<ObjectIdent>, CorbaImplCompatibility> get_ident_variants_from(const ObjectIdent& _object_ident);

} // namespace Fred::Logger::LogEntry

struct LogEntryPropertyChild
{
    LogEntryPropertyChild() = default;

    LogEntryPropertyChild(
            const LogEntry::PropertyType& _property_type,
            const LogEntry::PropertyValue& _property_value);

    LogEntry::PropertyType property_type;
    LogEntry::PropertyValue property_value;
};

struct LogEntryProperty
{
    LogEntryProperty() = default;

    LogEntryProperty(
            const LogEntry::PropertyValue& _property_value,
            const std::vector<LogEntryPropertyChild>& _children);

    LogEntry::PropertyValue property_value;
    std::vector<LogEntryPropertyChild> children;
};

struct LogEntryProperties
{
    LogEntryProperties() = default;

    LogEntryProperties(
            const LogEntry::PropertyType& _property_type,
            const std::vector<LogEntryProperty>& _properties);

    LogEntry::PropertyType property_type;
    std::vector<LogEntryProperty> properties;
};

struct ObjectReferences
{
    ObjectReferences() = default;

    ObjectReferences(
            const LogEntry::LogEntryObjectTypeName& _object_type,
            const std::vector<LogEntry::ObjectIdent>& _object_reference_ids);

    LogEntry::LogEntryObjectTypeName object_type;
    std::vector<LogEntry::ObjectIdent> object_reference_ids;
};

struct WrapperError : std::exception { };
struct UnwrapperError : std::exception { };

} // namespace Fred::Logger
} // namespace Fred

namespace boost {
namespace asio {
namespace ip {

std::string wrap_into_psql_representation(const boost::asio::ip::address& value);

boost::asio::ip::address unwrap_from_psql_representation(const boost::asio::ip::address&, const char* src);

} // namespace boost::asio::ip
} // namespace boost::asio
} // namespace boost

#endif
