/*
 * Copyright (C) 2020  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "test/fixture.hh"
#include "test/util.hh"

#include "src/exceptions.hh"
#include "src/get_log_entry_info.hh"

#include "libpg/util/strong_type_operators.hh"

#include <boost/optional/optional_io.hpp>
#include <boost/test/unit_test.hpp>

#include <string>
#include <map>
#include <stdexcept>

namespace Test {

BOOST_AUTO_TEST_SUITE(TestGetLogEntryInfo)

BOOST_FIXTURE_TEST_CASE(service_not_found, Util::AddFixture<Util::SupplyCtx<Fixture::HasLogEntryIdentWithNonexistentService>>)
{
    const Fixture::GetLogEntryInfoData get_log_entry_info_data{fixture.log_entry_ident_with_nonexistent_service};

    BOOST_CHECK_THROW(
            ::Fred::Logger::get_log_entry_info(
                    fixture.transaction,
                    get_log_entry_info_data.log_entry_ident),
            ::Fred::Logger::ServiceNotFound);
}

BOOST_FIXTURE_TEST_CASE(log_entry_does_not_exist, Util::AddFixture<Util::SupplyCtx<Fixture::HasNonexistentLogEntryIdent>>)
{
    const Fixture::GetLogEntryInfoData get_log_entry_info_data{fixture.nonexistent_log_entry_ident};

    BOOST_CHECK_THROW(
            ::Fred::Logger::get_log_entry_info(
                    fixture.transaction,
                    get_log_entry_info_data.log_entry_ident),
            ::Fred::Logger::LogEntryDoesNotExist);
}

BOOST_FIXTURE_TEST_CASE(log_entry_has_invalid_time_begin, Util::AddFixture<Util::SupplyCtx<Fixture::HasLogEntryIdentWithInvalidTimeBegin>>)
{
    {
        const Fixture::GetLogEntryInfoData get_log_entry_info_data{fixture.log_entry_ident_with_invalid_time_begin_not_a_date_time};

        BOOST_CHECK_THROW(
                ::Fred::Logger::get_log_entry_info(
                        fixture.transaction,
                        get_log_entry_info_data.log_entry_ident),
                std::exception);
    }
    {
        const Fixture::GetLogEntryInfoData get_log_entry_info_data{fixture.log_entry_ident_with_invalid_time_begin_pos_infin};

        BOOST_CHECK_THROW(
                ::Fred::Logger::get_log_entry_info(
                        fixture.transaction,
                        get_log_entry_info_data.log_entry_ident),
                std::exception);
    }
    {
        const Fixture::GetLogEntryInfoData get_log_entry_info_data{fixture.log_entry_ident_with_invalid_time_begin_neg_infin};

        BOOST_CHECK_THROW(
                ::Fred::Logger::get_log_entry_info(
                        fixture.transaction,
                        get_log_entry_info_data.log_entry_ident),
                std::exception);
    }
}

//namespace {
//
//Fred::Logger::LogEntry::LogEntryTypeName id_to_name(const Fred::Logger::LogEntry::LogEntryTypeId& _log_entry_type)
//{
//    std::map<Fred::Logger::LogEntry::LogEntryTypeId, Fred::Logger::LogEntry::LogEntryTypeName> log_entry_type_to_db_handle{
//        {Fred::Logger::LogEntry::LogEntryTypeId{500}, Fred::Logger::LogEntry::LogEntryTypeName{"DomainCheck"}}};
//    const auto search = log_entry_type_to_db_handle.find(_log_entry_type);
//    const auto found = search != log_entry_type_to_db_handle.end(); 
//    if (!found)
//    {
//        throw std::runtime_error{"unknown log_entry_type"};
//    }
//    return search->second;
//}
//
//} // namespace Test::{TestGetLogEntryInfo}::{anonymous}

BOOST_FIXTURE_TEST_CASE(ok, Util::AddFixture<Util::SupplyCtx<Fixture::HasLogEntryIdent>>)
{
    ::Fred::Logger::GetLogEntryInfoRequest get_log_entry_info_data{fixture.log_entry_ident};
    ::Fred::Logger::GetLogEntryInfoReply result;

    BOOST_CHECK_NO_THROW(
            result = ::Fred::Logger::get_log_entry_info(
                    fixture.transaction,
                    get_log_entry_info_data.log_entry_ident));

    BOOST_CHECK_EQUAL(get_log_entry_info_data.log_entry_ident.id, fixture.log_entry_001.existing_log_entry.id);
    BOOST_CHECK_EQUAL(result.log_entry_info.time_begin, get_log_entry_info_data.log_entry_ident.time_begin);
    BOOST_CHECK_EQUAL(result.log_entry_info.service_name, fixture.log_entry_001.existing_log_entry.service_name);
    BOOST_CHECK_EQUAL(result.log_entry_info.log_entry_type, fixture.log_entry_001.existing_log_entry.log_entry_type);
    BOOST_CHECK_EQUAL(result.log_entry_info.username, fixture.log_entry_001.existing_log_entry.username);
    BOOST_CHECK_EQUAL(result.log_entry_info.user_id, fixture.log_entry_001.existing_log_entry.user_id);
    BOOST_CHECK_EQUAL(result.log_entry_info.object_references.size(), 1);
    BOOST_CHECK_EQUAL(result.log_entry_info.object_references.at(0).object_reference_ids.size(), 1);
    BOOST_CHECK_EQUAL(result.log_entry_info.object_references.at(0).object_reference_ids.at(0), fixture.log_entry_001.existing_log_entry.object_ident);

}

BOOST_AUTO_TEST_SUITE_END()

} // namespace Test
