/*
 * Copyright (C) 2019-2020  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "test/fixture.hh"
#include "test/util.hh"

#include "src/exceptions.hh"
#include "src/list_log_entries.hh"
#include "src/log_entry_ident.hh"

#include "libpg/util/strong_type_operators.hh"

#include <boost/test/unit_test.hpp>

#include <boost/date_time/posix_time/posix_time.hpp>

#include <string>
#include <vector>

namespace Test {

namespace {
const Fred::Logger::LogEntry::ServiceName service_name{"EPP"};
const Fred::Logger::LogEntry::TimeEnd to_date{boost::posix_time::second_clock::local_time()};
const Fred::Logger::LogEntry::TimeBegin from_date{to_date.get() - boost::gregorian::days(10)};
constexpr unsigned int count_limit = 5;
const Fred::Logger::LogEntry::UserName username{"test_username"};
const Fred::Logger::LogEntry::UserId user_id{1};
const Fred::Logger::LogEntry::LogEntryObjectTypeName object_type{"domain"};
const Fred::Logger::LogEntry::ObjectIdent object_ident{"222"};
const std::vector<Fred::Logger::LogEntry::LogEntryTypeName> log_entry_types{
        Fred::Logger::LogEntry::LogEntryTypeName{"DomainCheck"},
        Fred::Logger::LogEntry::LogEntryTypeName{"DomainInfo"},
        Fred::Logger::LogEntry::LogEntryTypeName{"DomainDelete"},
        Fred::Logger::LogEntry::LogEntryTypeName{"DomainUpdate"},
        Fred::Logger::LogEntry::LogEntryTypeName{"DomainCreate"},
        Fred::Logger::LogEntry::LogEntryTypeName{"DomainTransfer"},
        Fred::Logger::LogEntry::LogEntryTypeName{"DomainRenew"},
        Fred::Logger::LogEntry::LogEntryTypeName{"DomainTrade"}};

bool operator>=(const Fred::Logger::LogEntry::TimeBegin& _lhs, const Fred::Logger::LogEntry::TimeEnd& _rhs)
{
    return _lhs.get() >= _rhs.get();
}

bool check_log_entries(
        const std::vector<Fred::Logger::LogEntryIdent>& _log_entries,
        const Fred::Logger::LogEntry::TimeBegin _from,
        const Fred::Logger::LogEntry::TimeEnd _to,
        const Fred::Logger::LogEntry::ServiceName& _service)
{
    for (const auto& log_entry : _log_entries)
    {
        if (log_entry.id.get() < 1)
        {
            throw std::runtime_error("Invalid id of log entry.");
        }
        if (log_entry.time_begin < _from || log_entry.time_begin >= _to)
        {
            throw std::runtime_error("Time begin has invalid range.");
        }
        if (log_entry.service_handle != _service)
        {
            throw std::runtime_error("Service name is " + log_entry.service_handle.get() + " instead of " + _service.get());
        }
        if (log_entry.monitoring_status != Fred::Logger::LogEntry::IsMonitoring{false})
        {
            throw std::runtime_error("Monitoring status is not false.");
        }
    }
    return true;
}

}// namespace Test::{anonymous}

BOOST_AUTO_TEST_SUITE(TestListLogEntries)

BOOST_FIXTURE_TEST_CASE(service_not_found, Util::AddFixture<Util::SupplyCtx<Fixture::Empty>>)
{
    Fixture::ListLogEntriesData list_log_entries_data;
    list_log_entries_data.service = Fred::Logger::LogEntry::ServiceName{"nonexistent_service"};
    list_log_entries_data.from = from_date;
    list_log_entries_data.to = to_date;
    list_log_entries_data.count_limit = count_limit;

    BOOST_CHECK_THROW(
            ::Fred::Logger::list_log_entries(
                    fixture.transaction,
                    list_log_entries_data.service,
                    list_log_entries_data.from,
                    list_log_entries_data.to,
                    list_log_entries_data.count_limit,
                    list_log_entries_data.username,
                    list_log_entries_data.user_id,
                    list_log_entries_data.object_reference,
                    list_log_entries_data.log_entry_types),
            ::Fred::Logger::ServiceNotFound);
}

BOOST_FIXTURE_TEST_CASE(bad_from_date, Util::AddFixture<Util::SupplyCtx<Fixture::Empty>>)
{
    ::Fred::Logger::ListLogEntriesRequest list_log_entries_data;
    list_log_entries_data.service = service_name;
    list_log_entries_data.to = to_date;
    list_log_entries_data.count_limit = count_limit;

    BOOST_CHECK_THROW(
            ::Fred::Logger::list_log_entries(
                    fixture.transaction,
                    list_log_entries_data.service,
                    list_log_entries_data.from,
                    list_log_entries_data.to,
                    list_log_entries_data.count_limit,
                    list_log_entries_data.username,
                    list_log_entries_data.user_id,
                    list_log_entries_data.object_reference,
                    list_log_entries_data.log_entry_types),
            ::Fred::Logger::InvalidTimestampFrom);
}

BOOST_FIXTURE_TEST_CASE(bad_to_date, Util::AddFixture<Util::SupplyCtx<Fixture::Empty>>)
{
    ::Fred::Logger::ListLogEntriesRequest list_log_entries_data;
    list_log_entries_data.service = service_name;
    list_log_entries_data.from = from_date;
    list_log_entries_data.count_limit = count_limit;

    BOOST_CHECK_THROW(
            ::Fred::Logger::list_log_entries(
                    fixture.transaction,
                    list_log_entries_data.service,
                    list_log_entries_data.from,
                    list_log_entries_data.to,
                    list_log_entries_data.count_limit,
                    list_log_entries_data.username,
                    list_log_entries_data.user_id,
                    list_log_entries_data.object_reference,
                    list_log_entries_data.log_entry_types),
            ::Fred::Logger::InvalidTimestampTo);
}

BOOST_FIXTURE_TEST_CASE(bad_date_range, Util::AddFixture<Util::SupplyCtx<Fixture::Empty>>)
{
    ::Fred::Logger::ListLogEntriesRequest list_log_entries_data;
    list_log_entries_data.service = service_name;
    list_log_entries_data.from = from_date;
    list_log_entries_data.to = Fred::Logger::LogEntry::TimeEnd{from_date.get()};
    list_log_entries_data.count_limit = count_limit;

    BOOST_CHECK_THROW(
            ::Fred::Logger::list_log_entries(
                    fixture.transaction,
                    list_log_entries_data.service,
                    list_log_entries_data.from,
                    list_log_entries_data.to,
                    list_log_entries_data.count_limit,
                    list_log_entries_data.username,
                    list_log_entries_data.user_id,
                    list_log_entries_data.object_reference,
                    list_log_entries_data.log_entry_types),
            ::Fred::Logger::InvalidTimestampRange);
}

BOOST_FIXTURE_TEST_CASE(bad_count_limit, Util::AddFixture<Util::SupplyCtx<Fixture::Empty>>)
{
    ::Fred::Logger::ListLogEntriesRequest list_log_entries_data;
    list_log_entries_data.service = service_name;
    list_log_entries_data.from = from_date;
    list_log_entries_data.to = to_date;
    list_log_entries_data.count_limit = 0;

    BOOST_CHECK_THROW(
            ::Fred::Logger::list_log_entries(
                    fixture.transaction,
                    list_log_entries_data.service,
                    list_log_entries_data.from,
                    list_log_entries_data.to,
                    list_log_entries_data.count_limit,
                    list_log_entries_data.username,
                    list_log_entries_data.user_id,
                    list_log_entries_data.object_reference,
                    list_log_entries_data.log_entry_types),
            ::Fred::Logger::InvalidCountLimit);
}

BOOST_FIXTURE_TEST_CASE(test_list_log_entries_empty_response, Util::AddFixture<Util::SupplyCtx<Fixture::HasSeveralLogEntries>>)
{
    ::Fred::Logger::ListLogEntriesRequest list_log_entries_data;
    list_log_entries_data.service = service_name;
    list_log_entries_data.from = from_date;
    list_log_entries_data.to = to_date;
    list_log_entries_data.count_limit = count_limit;
    list_log_entries_data.username = Fred::Logger::LogEntry::UserName{"nonexistent_username"};

    ::Fred::Logger::ListLogEntriesReply result;
    BOOST_CHECK_NO_THROW(
            result = ::Fred::Logger::list_log_entries(
                    fixture.transaction,
                    list_log_entries_data.service,
                    list_log_entries_data.from,
                    list_log_entries_data.to,
                    list_log_entries_data.count_limit,
                    list_log_entries_data.username,
                    list_log_entries_data.user_id,
                    list_log_entries_data.object_reference,
                    list_log_entries_data.log_entry_types));

    BOOST_CHECK(result.log_entry_ids.empty());
}

BOOST_FIXTURE_TEST_CASE(test_list_log_entries_empty_object_reference, Util::AddFixture<Util::SupplyCtx<Fixture::HasSeveralLogEntries>>)
{
    ::Fred::Logger::ListLogEntriesRequest list_log_entries_data;
    list_log_entries_data.service = service_name;
    list_log_entries_data.from = from_date;
    list_log_entries_data.to = to_date;
    list_log_entries_data.count_limit = count_limit;
    ::Fred::Logger::ObjectReference object_reference;
    object_reference.object_type = object_type;
    object_reference.object_ident = Fred::Logger::LogEntry::ObjectIdent{""};
    list_log_entries_data.object_reference = object_reference;
    list_log_entries_data.log_entry_types = log_entry_types;

    ::Fred::Logger::ListLogEntriesReply result;
    BOOST_CHECK_NO_THROW(
            result = ::Fred::Logger::list_log_entries(
                    fixture.transaction,
                    list_log_entries_data.service,
                    list_log_entries_data.from,
                    list_log_entries_data.to,
                    list_log_entries_data.count_limit,
                    list_log_entries_data.username,
                    list_log_entries_data.user_id,
                    list_log_entries_data.object_reference,
                    list_log_entries_data.log_entry_types));

    BOOST_CHECK(!result.log_entry_ids.empty());
    BOOST_CHECK(result.log_entry_ids.size() <= list_log_entries_data.count_limit);
    BOOST_CHECK(check_log_entries(result.log_entry_ids, from_date, to_date, service_name));
}

BOOST_FIXTURE_TEST_CASE(test_list_log_entries_count_limit, Util::AddFixture<Util::SupplyCtx<Fixture::HasSeveralLogEntries>>)
{
    ::Fred::Logger::ListLogEntriesRequest list_log_entries_data;
    list_log_entries_data.service = service_name;
    list_log_entries_data.from = from_date;
    list_log_entries_data.to = to_date;
    list_log_entries_data.count_limit = 3;

    ::Fred::Logger::ListLogEntriesReply result;
    BOOST_CHECK_NO_THROW(
            result = ::Fred::Logger::list_log_entries(
                    fixture.transaction,
                    list_log_entries_data.service,
                    list_log_entries_data.from,
                    list_log_entries_data.to,
                    list_log_entries_data.count_limit,
                    list_log_entries_data.username,
                    list_log_entries_data.user_id,
                    list_log_entries_data.object_reference,
                    list_log_entries_data.log_entry_types));

    BOOST_CHECK(result.log_entry_ids.size() == list_log_entries_data.count_limit);
    BOOST_CHECK(check_log_entries(result.log_entry_ids, from_date, to_date, service_name));
}

BOOST_FIXTURE_TEST_CASE(test_list_log_entries_min, Util::AddFixture<Util::SupplyCtx<Fixture::HasSeveralLogEntries>>)
{
    ::Fred::Logger::ListLogEntriesRequest list_log_entries_data;
    list_log_entries_data.service = service_name;
    list_log_entries_data.from = from_date;
    list_log_entries_data.to = to_date;
    list_log_entries_data.count_limit = count_limit;

    ::Fred::Logger::ListLogEntriesReply result;
    BOOST_CHECK_NO_THROW(
            result = ::Fred::Logger::list_log_entries(
                    fixture.transaction,
                    list_log_entries_data.service,
                    list_log_entries_data.from,
                    list_log_entries_data.to,
                    list_log_entries_data.count_limit,
                    list_log_entries_data.username,
                    list_log_entries_data.user_id,
                    list_log_entries_data.object_reference,
                    list_log_entries_data.log_entry_types));

    BOOST_CHECK(!result.log_entry_ids.empty());
    BOOST_CHECK(result.log_entry_ids.size() <= list_log_entries_data.count_limit);
    BOOST_CHECK(check_log_entries(result.log_entry_ids, from_date, to_date, service_name));
}

BOOST_FIXTURE_TEST_CASE(test_list_log_entries_max, Util::AddFixture<Util::SupplyCtx<Fixture::HasSeveralLogEntries>>)
{
    ::Fred::Logger::ListLogEntriesRequest list_log_entries_data;
    list_log_entries_data.service = service_name;
    list_log_entries_data.from = from_date;
    list_log_entries_data.to = to_date;
    list_log_entries_data.count_limit = count_limit;

    list_log_entries_data.username = username;
    list_log_entries_data.user_id = user_id;
    ::Fred::Logger::ObjectReference object_reference;
    object_reference.object_type = object_type;
    object_reference.object_ident = object_ident;
    list_log_entries_data.object_reference = object_reference;
    list_log_entries_data.log_entry_types = log_entry_types;

    ::Fred::Logger::ListLogEntriesReply result;
    BOOST_CHECK_NO_THROW(
            result = ::Fred::Logger::list_log_entries(
                    fixture.transaction,
                    list_log_entries_data.service,
                    list_log_entries_data.from,
                    list_log_entries_data.to,
                    list_log_entries_data.count_limit,
                    list_log_entries_data.username,
                    list_log_entries_data.user_id,
                    list_log_entries_data.object_reference,
                    list_log_entries_data.log_entry_types));

    BOOST_CHECK(!result.log_entry_ids.empty());
    BOOST_CHECK(result.log_entry_ids.size() <= list_log_entries_data.count_limit);
    BOOST_CHECK(check_log_entries(result.log_entry_ids, from_date, to_date, service_name));
}

BOOST_AUTO_TEST_SUITE_END() // TestListLogEntries

} // namespace Test
